//  FiveQueueScheduler.cpp
//  Scheduler Simulator
//
//  Created by Wade Tregaskis on Sat May 01 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "FiveQueueScheduler.h"


void FiveQueueScheduler::updateInteractiveDisplay(void) {
    if (active) {
        static int linesToDelete = 0;
        struct timeval currentTime;
        
        //printf("\033[H\033[J"); // This moves home & clears the whole screen.  It doesn't work for buffered terminals however, as the clear screen command just pages down with whitespace - so you end up with a giant buffer full of every screen.
        
        //printf("\033[H"); // Just seeks home... unnecessary if the code below is correct
        
        while (linesToDelete > 0) {
            printf("\033[A\033[2K");
            --linesToDelete;
        }
        
        if (0 != gettimeofday(&currentTime, NULL)) {
            fprintf(stderr, "FiveQueueScheduler::updateInteractiveDisplay() - Unable to get time of day.\n"); ++linesToDelete;
        } else {
            unsigned long toRunSize = toRun.size(), readySize = ready.size(), blockedSize = blocked.size(), finishedSize = finished.size(), executingSize = executing.size(); // It's important to only call size() on a list once, where possible, as calculating the size involves iterating over the entire contents of the list.  This is incredibly slow for more than a few hundred thousand entries.
            ProcessInfo *current;
            list<ProcessInfo*>::const_iterator iter, stopper;
            double averagePriority = 0.0, averageResponse = 0.0, averageTurnaround = 0.0, averageWaitTime = 0.0, averageExecution = 0.0, averageRemainingExecution = 0.0, averageWResponse = 0.0, averageWTurnaround = 0.0, averageWWaitTime = 0.0;
            double curResponse, curTurnaround, curWaitTime;
            
            //printf("FCFS Scheduler - %lu process%s to run, %lu process%s ready, %lu process%s blocked, %lu process%s executing, %lu process%s finished.\n", toRunSize, ((toRunSize == 1) ? "" : "es"), readySize, ((readySize == 1) ? "" : "es"), blockedSize, ((blockedSize == 1) ? "" : "es"), executingSize, ((executingSize == 1) ? "" : "es"), finishedSize, ((finishedSize == 1) ? "" : "es")); ++linesToDelete;
            
            printf("Scheduled %d quantums (%.2f seconds) since activation at %s (currently ", ticksSoFar, ticksSoFar * currentTimeQuantum, timeAsString(lastActivated));
            printf("%s).\n", timeAsString(currentTime)); ++linesToDelete;
            printf("%lu to run, %lu ready, %lu blocked, %lu running, %lu finished.\n", toRunSize, readySize, blockedSize, executingSize, finishedSize); ++linesToDelete;
            
            printf("+----------+-------+-------+--------+--------+--------+--------+--------+\n"); ++linesToDelete;
            printf("|  QUEUE   |  PID  | Prior | R Time | T Time | E Time | E Rema | W Time |\n"); ++linesToDelete;
            printf("+----------+-------+-------+--------+--------+--------+--------+--------+\n"); ++linesToDelete;
            
            if (toRunSize > 0) {
                printf("| To Run   |       |       |        |        |        |        |        |\n"); ++linesToDelete;
                
                for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    averagePriority += current->currentPriority;
                        
                    curResponse = timeDifference(currentTime, current->launched);
                    curTurnaround = timeDifference(currentTime, current->launched);
                    curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);
                    
                    averageResponse += curResponse;
                    averageWResponse += (1.0 / current->priority) * curResponse;
                    averageTurnaround += curTurnaround;
                    averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                    averageWaitTime += curWaitTime;
                    averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                    averageExecution += current->executionTime;
                    averageRemainingExecution += current->executionTimeRemaining;
                    
                    printf("|          |       | %5.3f | +%5.2f | +%5.2f | %6.2f | %6.2f | +%5.2f |\n", current->priority, curResponse, curTurnaround, current->executionTime, current->executionTimeRemaining, curWaitTime); ++linesToDelete;
                }
            }
            
            if (readySize > 0) {
                printf("| Ready    |       |       |        |        |        |        |        |\n"); ++linesToDelete;
                
                for (iter = ready.begin(), stopper = ready.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    averagePriority += current->currentPriority;
                    
                    curResponse = timeDifference(current->firstServed, current->launched);
                    curTurnaround = timeDifference(currentTime, current->launched);
                    curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);

                    averageResponse += curResponse;
                    averageWResponse += (1.0 / current->priority) * curResponse;
                    averageTurnaround += curTurnaround;
                    averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                    averageWaitTime += curWaitTime;
                    averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                    averageExecution += current->executionTime;
                    averageRemainingExecution += current->executionTimeRemaining;

                    printf("|          | %5d | %5.3f | %6.2f | +%5.2f | %6.2f | %6.2f | +%5.2f |\n", current->pid, current->currentPriority, curResponse, curTurnaround, current->executionTime, current->executionTimeRemaining, curWaitTime); ++linesToDelete;
                }
            }
            
            if (blockedSize > 0) {
                printf("| Blocked  |       |       |        |        |        |        |        |\n"); ++linesToDelete;
                
                for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    averagePriority += current->currentPriority;
                    
                    curResponse = timeDifference(current->firstServed, current->launched);
                    curTurnaround = timeDifference(currentTime, current->launched);
                    curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);

                    averageResponse += curResponse;
                    averageWResponse += (1.0 / current->priority) * curResponse;
                    averageTurnaround += curTurnaround;
                    averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                    averageWaitTime += curWaitTime;
                    averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                    averageExecution += current->executionTime;
                    averageRemainingExecution += current->executionTimeRemaining;

                    printf("|          | %5d | %5.3f | %6.2f | +%5.2f | %6.2f | %6.2f | +%5.2f |\n", current->pid, current->currentPriority, curResponse, curTurnaround, current->executionTime, current->executionTimeRemaining, curWaitTime); ++linesToDelete;
                }
            }
            
            if (executingSize > 0) {
                printf("| Running  |       |       |        |        |        |        |        |\n"); ++linesToDelete;
                
                for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    averagePriority += current->currentPriority;
                    
                    curResponse = timeDifference(current->firstServed, current->launched);
                    curTurnaround = timeDifference(currentTime, current->launched);
                    curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);

                    averageResponse += curResponse;
                    averageWResponse += (1.0 / current->priority) * curResponse;
                    averageTurnaround += curTurnaround;
                    averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                    averageWaitTime += curWaitTime;
                    averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                    averageExecution += current->executionTime;
                    averageRemainingExecution += current->executionTimeRemaining;

                    printf("|          | %5d | %5.3f | %6.2f | +%5.2f | %6.2f | %6.2f | +%5.2f |\n", current->pid, current->currentPriority, curResponse, curTurnaround, current->executionTime, current->executionTimeRemaining, curWaitTime); ++linesToDelete;
                }
            }
            
            if (finishedSize > 0) {
                printf("| Finished |       |       |        |        |        |        |        |\n"); ++linesToDelete;
                
                for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    averagePriority += current->currentPriority;
                    
                    curResponse = timeDifference(current->firstServed, current->launched);
                    curTurnaround = timeDifference(current->ended, current->launched);
                    curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);

                    averageResponse += curResponse;
                    averageWResponse += (1.0 / current->priority) * curResponse;
                    averageTurnaround += curTurnaround;
                    averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                    averageWaitTime += curWaitTime;
                    averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                    averageExecution += current->executionTime;
                    
                    printf("|          | %5d | %5.3f | %6.2f | %6.2f | %6.2f |        | %6.2f |\n", current->pid, current->currentPriority, curResponse, curTurnaround, current->executionTime, curWaitTime); ++linesToDelete;
                }
            }
            
            double fullCount = toRunSize + readySize + blockedSize + executingSize + finishedSize;
            
            averagePriority /= fullCount;
            averageResponse /= fullCount;
            averageWResponse /= fullCount;
            averageTurnaround /= fullCount;
            averageWTurnaround /= fullCount;
            averageWaitTime /= fullCount;
            averageWWaitTime /= fullCount;
            averageExecution /= fullCount;
            averageRemainingExecution /= fullCount;
            
            printf("+----------+-------+-------+--------+--------+--------+--------+--------+\n"); ++linesToDelete;
            
            printf("| AVERAGE  |       | %5.2f | %6.2f | %6.2f | %6.2f | %6.2f | %6.2f |\n", averagePriority, averageResponse, averageTurnaround, averageExecution, averageRemainingExecution, averageWaitTime); ++linesToDelete;
            printf("| WAVERAGE |       |       | %6.2f | %6.2f |        |        | %6.2f |\n", averageWResponse, averageWTurnaround, averageWWaitTime); ++linesToDelete;

            printf("+----------+-------+-------+--------+--------+--------+--------+--------+\n"); ++linesToDelete;
            
            fflush(stdout); // There seems to be some issues with buffering and terminal control... hopefully flushing explicitly will prevent these arising
        }
    }
}

FiveQueueScheduler::~FiveQueueScheduler(void) {
    list<ProcessInfo*>::iterator iter, stopper;
    ProcessInfo *current;
    
    for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
        delete *iter;
    }
    
    for (iter = ready.begin(), stopper = ready.end(); iter != stopper; ++iter) {
        current = *iter;
        
        if (0 != kill(current->pid, SIGKILL)) {
            fprintf(stderr, "FiveQueueScheduler::~FiveQueueScheduler() - Unable to kill process #%d, error #%d.\n", current->pid, errno);
        }
        
        delete current;
    }
    
    for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
        current = *iter;
        
        if (0 != kill(current->pid, SIGKILL)) {
            fprintf(stderr, "FiveQueueScheduler::~FiveQueueScheduler() - Unable to kill process #%d, error #%d.\n", current->pid, errno);
        }
        
        delete current;
    }
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        current = *iter;
        
        if (0 != kill(current->pid, SIGKILL)) {
            fprintf(stderr, "FiveQueueScheduler::~FiveQueueScheduler() - Unable to kill process #%d, error #%d.\n", current->pid, errno);
        }
        
        delete current;
    }
    
    for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
        delete *iter;
    }
}

void FiveQueueScheduler::setActive(bool beActive) {
    active = beActive;
    
    if (active) {
        list<ProcessInfo*>::iterator iter, stopper;
        ProcessInfo *current;
        
        gettimeofday(&lastActivated, NULL);
        
        for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
            current = *iter;
            
            if ((0 == current->launched.tv_sec) && (0 == current->launched.tv_usec)) {
                gettimeofday(&(current->launched), NULL);
            }
        }
        
        if (executing.empty()) {
            sigset_t oldSignalSet, blockSet;
            
            sigemptyset(&blockSet);
            sigaddset(&blockSet, SIGUSR1);
            sigaddset(&blockSet, SIGCHLD);
            sigaddset(&blockSet, SIGALRM);
            
            assert(sigprocmask(SIG_BLOCK, &blockSet, &oldSignalSet) == 0); // We're not necessarily running in a signal handler, therefore we must make sure to block the relevant signals to avoid having them called during the process management in runNextProcess(); failing to do so can result in us actually SIGSTOP'ing ourselves, because the currently executing child, if coming from the toRun queue, might not get it's PID set appropriately (and thus remain defaulted to 0) after calling spawnChild() if a signal occurs (e.g. SIGALRM) during spawnChild.  Passing 0 to kill() sends it to ourselves.  D'oh.
            ++ticksSoFar;
            this->runNextProcess();
            assert(sigprocmask(SIG_SETMASK, &oldSignalSet, NULL) == 0);
        }
    }
}

void FiveQueueScheduler::tick(void) {
    ++ticksSoFar;
    
    ready.insert(ready.end(), blocked.begin(), blocked.end());
    blocked.clear();
    
    if (executing.empty()) {
        this->runNextProcess();
    } else {
        list<ProcessInfo*>::iterator iter, stopper;
        
        for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
            (*iter)->executionTimeRemaining -= currentTimeQuantum;
        }
    }
    
    if (interactive) {
        this->updateInteractiveDisplay();
    }
}

void FiveQueueScheduler::startProcess(double priority, double executionTime) {
    ProcessInfo *temp = new ProcessInfo;
    
    temp->priority = temp->currentPriority = priority;
    temp->executionTime = temp->executionTimeRemaining = executionTime;
    
    if (active) {
        gettimeofday(&(temp->launched), NULL);
    } else {
        temp->launched.tv_sec = temp->launched.tv_usec = 0;
    }
    
    toRun.push_back(temp);
    
    if (executing.empty()) {
        this->runNextProcess();
    }
    
    if (interactive) {
        this->updateInteractiveDisplay();
    }
}

vector<pid_t>* FiveQueueScheduler::processList(void) {
    list<ProcessInfo*>::const_iterator iter, stopper;
    vector<pid_t> *result = new vector<pid_t>;
    
    assert(result);
    
    for (iter = ready.begin(), stopper = ready.end(); iter != stopper; ++iter) {
        result->push_back((*iter)->pid);
    }
    
    for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
        result->push_back((*iter)->pid);
    }
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        result->push_back((*iter)->pid);
    }
    
    return result;
}

ProcessInfo* FiveQueueScheduler::processInfo(pid_t pid) {
    list<ProcessInfo*>::const_iterator iter, stopper;
    
    for (iter = ready.begin(), stopper = ready.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            return copyInfo(*iter);
        }
    }
    
    for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            return copyInfo(*iter);
        }
    }
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            return copyInfo(*iter);
        }
    }
    
    for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            return copyInfo(*iter);
        }
    }
    
    return NULL;
}

void FiveQueueScheduler::killProcess(pid_t pid) {
    list<ProcessInfo*>::const_iterator iter, stopper;
    
    for (iter = ready.begin(), stopper = ready.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            kill(SIGKILL, pid); return;
        }
    }
    
    for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            kill(SIGKILL, pid); return;
        }
    }
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            kill(SIGKILL, pid); return;
        }
    }
    
    for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            kill(SIGKILL, pid); return;
        }
    }
    
    fprintf(stderr, "FiveQueueScheduler::killProcess(%d) - Unable to find any process with that real-world PID.  Remember I can only kill processes that I am managing, not just any arbitrary ones.\n", pid);
}

void FiveQueueScheduler::processBlocked(pid_t pid) {
    list<ProcessInfo*>::iterator iter, stopper;
    ProcessInfo *current;
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        current = *iter;
        
        blocked.push_back(current);
        executing.erase(iter);
        
        return;
    }
    
    if (current) {
        this->runNextProcess();
        
        if (interactive) {
            this->updateInteractiveDisplay();
        }
    } else {
        fprintf(stderr, "FiveQueueScheduler::processBlocked(%d) - Cannot find an executing process with the given PID.\n", pid);
    }
}

void FiveQueueScheduler::processTerminated(pid_t pid, int status) {
    list<ProcessInfo*>::iterator iter, stopper;
    ProcessInfo *selection = NULL, *current;
    bool wasActive = false;
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        current = *iter;
        
        if (current->pid == pid) {
            selection = current;
            executing.erase(iter);
            wasActive = true;
            break;
        }
    }
    
    if (!selection) {
        for (iter = ready.begin(), stopper = ready.end(); iter != stopper; ++iter) {
            current = *iter;
            
            if (current->pid == pid) {
                selection = current;
                ready.erase(iter);
                break;
            }
        }
        
        if (!selection) {
            for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
                current = *iter;
                
                if (current->pid == pid) {
                    selection = current;
                    blocked.erase(iter);
                    break;
                }
            }
            
            if (!selection) {
                fprintf(stderr, "FiveQueueScheduler::processTerminated(%d, %d) - Unable to find any record of the given process.\n", pid, status);
                return;
            }
        }
    }
    
    selection->result = status;
    
    if (0 != gettimeofday(&(selection->ended), NULL)) {
        fprintf(stderr, "FiveQueueScheduler::processTerminated() - Unable to get time of day, error #%d.\n", errno);
    }
    
    finished.push_back(selection);
    
    if (wasActive) { // Due to some timing issues, we might change processes but then find that the one we just 'paused' has in fact finished (and was awaiting processing behind the ALRM signal).  In this case, we don't want to run a replacement process for the dead one, since we already have something else running.  So, we only run another process if the one that just died was active.  Got it? :)
        this->runNextProcess();
    }
    
    if (interactive) {
        this->updateInteractiveDisplay();
    }
}

void FiveQueueScheduler::outputState(ostream &out) {
    unsigned long toRunSize = toRun.size(), readySize = ready.size(), blockedSize = blocked.size(), finishedSize = finished.size(), executingSize = executing.size(); // It's important to only call size() on a list once, where possible, as calculating the size involves iterating over the entire contents of the list.  This is incredibly slow for more than a few hundred thousand entries.
    ProcessInfo *current;
    list<ProcessInfo*>::const_iterator iter, stopper;
    
    out << toRunSize << " process" << ((toRunSize == 1) ? "" : "es") << " to run, " << readySize << " process" << ((readySize == 1) ? "" : "es") << " ready, " << blockedSize << " process" << ((blockedSize == 1) ? "" : "es") << " blocked, " << executingSize << " process" << ((executingSize == 1) ? "" : "es") << " executing, " << finishedSize << " process" << ((finishedSize == 1) ? "" : "es") << " finished.\n";
    
    if (toRunSize > 0) {
        out << "To run:\n";
        
        for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
            current = *iter;
            
            printf("\tPriority: %2.2f\tExecution time: %2.2f\tLaunched: %s (%s)\n", current->priority, current->executionTime, timeAsString(current->launched), timeSinceNowAsString(current->launched));
        }
    }
    
    if (readySize > 0) {
        out << "Ready:\n";
        
        for (iter = ready.begin(), stopper = ready.end(); iter != stopper; ++iter) {
            current = *iter;
            
            out << "\tPriority: " << current->priority << "\tExecution time: " << current->executionTime << "\tLaunched: " << timeAsString(current->launched) << " (" << timeSinceNowAsString(current->launched) << ")\tPID: " << current->pid << "\tCurrent priority: " << current->currentPriority;
            out << "\tFirst served: " << timeAsString(current->firstServed) << " (" << timeSinceNowAsString(current->firstServed) << ")\n";
        }
    }
    
    if (blockedSize > 0) {
        out << "Blocked:\n";
        
        for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
            current = *iter;
            
            out << "\tPriority: " << current->priority << "\tExecution time: " << current->executionTime << "\tLaunched: " << timeAsString(current->launched) << " (" << timeSinceNowAsString(current->launched) << ")\tPID: " << current->pid << "\tCurrent priority: " << current->currentPriority;
            out << "\tFirst served: " << timeAsString(current->firstServed) << " (" << timeSinceNowAsString(current->firstServed) << ")\n";}
    }
    
    if (executingSize > 0) {
        out << "Executing:\n";
        
        for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
            current = *iter;
            
            out << "\tPriority: " << current->priority << "\tExecution time: " << current->executionTime << "\tLaunched: " << timeAsString(current->launched) << " (" << timeSinceNowAsString(current->launched) << ")\tPID: " << current->pid << "\tCurrent priority: " << current->currentPriority;
            out << "\tFirst served: " << timeAsString(current->firstServed) << " (" << timeSinceNowAsString(current->firstServed) << ")\n";
        }
    }
    
    if (finishedSize > 0) {
        out << "Finished:\n";
        
        for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
            current = *iter;
            
            out << "\tPriority: " << current->priority << "\tExecution time: " << current->executionTime << "\tLaunched: " << timeAsString(current->launched) << " (" << timeSinceNowAsString(current->launched) << ")\tPID: " << current->pid << "\tCurrent priority: " << current->currentPriority;
            out << "\tFirst served: " << timeAsString(current->firstServed) << " (" << timeSinceNowAsString(current->firstServed) << ")";
            out << "\tEnded: " << timeAsString(current->ended) << " (" << timeSinceNowAsString(current->ended) << ")\n";
        }
    }
}

void FiveQueueScheduler::outputPerformance(ostream &out) {
    list<ProcessInfo*>::const_iterator iter, stopper;
    ProcessInfo *current;
    double bestTurnaround = 10000000000.0, totalTurnaround = 0.0, worstTurnaround = 0.0, bestResponse = 10000000000.0, totalResponse = 0.0, worstResponse = 0.0, currentTurnaround, currentResponse; // Note that the best in both cases is set to a massive initial value to ensure they will take on the value of the first process considered.  For assurance, the given defaults represent a little over 3 centuries of time.. if you run this program for that long, good for you.
    double bestTurnaroundWeighted = 10000000000.0, totalTurnaroundWeighted = 0.0, worstTurnaroundWeighted = 0.0, bestResponseWeighted = 10000000000.0, totalResponseWeighted = 0.0, worstResponseWeighted = 0.0, currentTurnaroundWeighted, currentResponseWeighted;
    
    int count = 0;
    
    if (!finished.empty()) {
        for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
            ++count;
            current = *iter;
            
            currentTurnaround = (current->ended.tv_sec + (current->ended.tv_usec / 1000000.0)) - (current->launched.tv_sec + (current->launched.tv_usec / 1000000.0));
            currentResponse = (current->firstServed.tv_sec + (current->firstServed.tv_usec / 1000000.0)) - (current->launched.tv_sec + (current->launched.tv_usec / 1000000.0));
            
            currentTurnaroundWeighted = currentTurnaround * (1.0 / current->priority);
            currentResponseWeighted = currentResponse * (1.0 / current->priority);
            
            totalTurnaround += currentTurnaround;
            totalResponse += currentResponse;
            
            totalTurnaroundWeighted += currentTurnaroundWeighted;
            totalResponseWeighted += currentResponseWeighted;
            
            if (currentTurnaround < bestTurnaround) {
                bestTurnaround = currentTurnaround;
            }
            
            if (currentTurnaround > worstTurnaround) {
                worstTurnaround = currentTurnaround;
            }
            
            if (currentResponse < bestResponse) {
                bestResponse = currentResponse;
            }
            
            if (currentResponse > worstResponse) {
                worstResponse = currentResponse;
            }
            
            if (currentTurnaroundWeighted < bestTurnaroundWeighted) {
                bestTurnaroundWeighted = currentTurnaroundWeighted;
            }
            
            if (currentTurnaroundWeighted > worstTurnaroundWeighted) {
                worstTurnaroundWeighted = currentTurnaroundWeighted;
            }
            
            if (currentResponseWeighted < bestResponseWeighted) {
                bestResponseWeighted = currentResponseWeighted;
            }
            
            if (currentResponseWeighted > worstResponseWeighted) {
                worstResponseWeighted = currentResponseWeighted;
            }
        }
        
        out << count << " process" << ((count == 1) ? "" : "es") << " completed:\n\tTurnaround { Best: " << bestTurnaround << "\tWorst: " << worstTurnaround << "\tAverage: " << (totalTurnaround / count) << " }\n\tResponse { Best: " << bestResponse << "\tWorst: " << worstResponse << "\tAverage: " << (totalResponse / count) << " }\n\tWeighted turnaround { Best: " << bestTurnaroundWeighted << "\tWorst: " << worstTurnaroundWeighted << "\tAverage: " << (totalTurnaroundWeighted / count) << " }\n\tWeighted response { Best: " << bestResponseWeighted << "\tWorst: " << worstResponseWeighted << "\tAverage: " << (totalResponseWeighted / count) << " }\n";
    }
}
