//  FiveQueueScheduler.h
//  Scheduler Simulator
//
//  Created by Wade Tregaskis on Sat May 01 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef FIVE_QUEUE_SCHEDULER_H
#define FIVE_QUEUE_SCHEDULER_H

#include "Scheduler.h"

#include <list>
using namespace std;


/*! @class FiveQueueScheduler
    @abstract Abstract base class for schedulers that use the standard five-queue model.
    @discussion Most simple schedulers use a 5-queue model: to-run, ready, blocked, executing & finished.  This class provides the base for this model.  Most schedulers using this model need only override the runNextProcess member function to implement the appropriate scheduling mechanism.
    */

class FiveQueueScheduler : public Scheduler {
protected:
    list<ProcessInfo*> toRun, ready, blocked, finished, executing; // The five queues
    
/*! @function runNextProcess
    @abstract Instructs the receiver to locate the next process to run, and do so.
    @discussion This is where most of the guts of the scheduler is; determining which process to run next.  This method is expected to choose some appropriate process and run it. */
    
    virtual void runNextProcess(void) = 0;

/*! @function updateInteractiveDisplay
    @abstract Updates the interactive display of scheduler operation.
    @discussion This method should display a table of the receiver's scheduled processes, alongside any additional information that is relevant.  It should also make sure to clear/overwrite the previous table, if one has been rendered.  It is allowed to assume that no other output is presented to the user inbetween invocations. */
    
    virtual void updateInteractiveDisplay(void);
    
public:
    FiveQueueScheduler(bool inter = false) : Scheduler(inter) {}
    ~FiveQueueScheduler();
    
    // Scheduler
    
    virtual void setActive(bool beActive);
    
    virtual void tick(void);
    
    virtual void startProcess(double priority, double executionTime);
    virtual vector<pid_t>* processList(void);
    virtual ProcessInfo* processInfo(pid_t pid);
    virtual void killProcess(pid_t pid);
    
    virtual void processBlocked(pid_t pid);
    virtual void processTerminated(pid_t pid, int status);

    virtual void outputState(ostream &out);
    virtual void outputPerformance(ostream &out);
};

#endif
