//  MLFQScheduler.cpp
//  Scheduler Simulator
//
//  Created by Wade Tregaskis on Fri Apr 09 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "MLFQScheduler.h"


static const int MLFQ_QUEUE_TYPE_MIN = 1;
const int MLFQ_RR = 1;
const int MLFQ_PRR = 2;
const int MLFQ_FCFS = 3;
const int MLFQ_SRJF = 4;
const int MLFQ_P = 5;
static const int MLFQ_QUEUE_TYPE_MAX = 5;


void MLFQScheduler::runNextProcess(void) {
    if (active) {
        ProcessInfo *current;
        
        if (((queueTypes[0] == MLFQ_SRJF) || (queueTypes[0] == MLFQ_P)) && (!toRun.empty() || !ready[0].empty())) {
            list<ProcessInfo*>::iterator iter, stopper;
            ProcessInfo *selection = NULL;
            bool toRunSource = true;
            
            if (queueTypes[0] == MLFQ_SRJF) {
                for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    if (!selection || (selection->executionTimeRemaining > current->executionTimeRemaining)) {
                        selection = current;
                    }
                }
                
                for (iter = ready[0].begin(), stopper = ready[0].end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    if (!selection || (selection->executionTimeRemaining > current->executionTimeRemaining)) {
                        selection = current;
                        toRunSource = false;
                    }
                }
            } else { // == MLFQ_P
                for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    if (!selection || (selection->currentPriority > current->currentPriority)) {
                        selection = current;
                    }
                }
                
                for (iter = ready[0].begin(), stopper = ready[0].end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    if (!selection || (selection->currentPriority > current->currentPriority)) {
                        selection = current;
                        toRunSource = false;
                    }
                }
            }
            
            remove((toRunSource ? &toRun : &ready[0]), selection);
            
            executing.push_back(selection);
            
            if (toRunSource) {
                if (0 != gettimeofday(&(selection->firstServed), NULL)) {
                    fprintf(stderr, "Warning: unable to get time of day, error #%d.\n", errno);
                }
                
                selection->queue = 0;
                
                selection->pid = spawnChild(selection->executionTime);
                
                if (selection->pid <= 0) {
                    fprintf(stderr, "Unable to spawn process, error #%d.\n", errno);
                }
                
                if (timeRemaining[selection->pid] == 0) {
                    timeRemaining[selection->pid] = timeQuantums[0];
                }
            } else {
                if (0 != kill(current->pid, SIGCONT)) {
                    fprintf(stderr, "Unable to wake up child process #%d - error #%d.\n", current->pid, errno);
                }
            }
        } else if (!toRun.empty() && (ready[0].empty() || (queueTypes[0] == MLFQ_RR) || (queueTypes[0] == MLFQ_PRR))) {
            if (queueTypes[0] == MLFQ_PRR) {
                list<ProcessInfo*>::iterator iter, stopper;
                
                current = NULL;
                
                for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
                    if (!current || (current->priority > (*iter)->priority)) {
                        current = *iter;
                    }
                }
                
                remove(&toRun, current);
            } else {
                current = toRun.front();
                toRun.pop_front();
            }
            
            executing.push_back(current);
                        
            if (0 != gettimeofday(&(current->firstServed), NULL)) {
                fprintf(stderr, "MLFQScheduler::runNextProcess() - Unable to get time of day, error #%d.\n", errno);
            }
            
            current->queue = 0;
            
            current->pid = spawnChild(current->executionTime);
            
            if (current->pid <= 0) {
                fprintf(stderr, "MLFQScheduler::runNextProcess() - Unable to spawn process, error #%d.\n", errno);                
            }
            
            if (timeRemaining[current->pid] == 0) {
                timeRemaining[current->pid] = timeQuantums[0];
            }
        } else {
            for (int i = 0; i < numberOfQueues; ++i) {
                if (!ready[i].empty()) {
                    ProcessInfo *selection = NULL;

                    if (queueTypes[i] == MLFQ_SRJF) {
                        list<ProcessInfo*>::iterator iter, stopper;
                        
                        for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
                            current = *iter;
                            
                            if (!selection || (selection->executionTimeRemaining > current->executionTimeRemaining)) {
                                selection = current;
                            }
                        }
                        
                        remove(&ready[i], selection);
                    } else if (queueTypes[i] == MLFQ_P) {
                        list<ProcessInfo*>::iterator iter, stopper;
                        
                        for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
                            current = *iter;
                            
                            if (!selection || (selection->currentPriority > current->currentPriority)) {
                                selection = current;
                            }
                        }
                        
                        remove(&ready[i], selection);
                    } else {
                        selection = ready[i].front();
                        ready[i].pop_front();
                        
                        if (timeRemaining[selection->pid] == 0) {
                            timeRemaining[selection->pid] = timeQuantums[i];
                        }
                    }
                    
                    executing.push_back(selection);
                    
                    if (0 != kill(selection->pid, SIGCONT)) {
                        fprintf(stderr, "MLFQScheduler::runNextProcess() - Unable to wake up child process #%d - error #%d.\n", selection->pid, errno);
                    }
                    
                    break;
                }
            }
        }
    }
}

void MLFQScheduler::updateInteractiveDisplay(void) {
    if (active) {
        static int linesToDelete = 0;
        struct timeval currentTime;
        
        
        while (linesToDelete > 0) {
            printf("\033[A\033[2K");
            --linesToDelete;
        }
        
        if (0 != gettimeofday(&currentTime, NULL)) {
            fprintf(stderr, "FiveQueueScheduler::updateInteractiveDisplay() - Unable to get time of day.\n"); ++linesToDelete;
        } else {
            unsigned long toRunSize = toRun.size(), readySize = 0, blockedSize = blocked.size(), finishedSize = finished.size(), executingSize = executing.size(); // It's important to only call size() on a list once, where possible, as calculating the size involves iterating over the entire contents of the list.  This is incredibly slow for more than a few hundred thousand entries.
            ProcessInfo *current;
            list<ProcessInfo*>::const_iterator iter, stopper;
            int i;
            double averagePriority = 0.0, averageResponse = 0.0, averageTurnaround = 0.0, averageWaitTime = 0.0, averageExecution = 0.0, averageRemainingExecution = 0.0, averageWResponse = 0.0, averageWTurnaround = 0.0, averageWWaitTime = 0.0;
            double curResponse, curTurnaround, curWaitTime;
            
            for (i = 0; i < numberOfQueues; ++i) {
                readySize += ready[i].size();
            }
            
            printf("Scheduled %d quantums (%.2f seconds) since activation at %s (currently ", ticksSoFar, ticksSoFar * currentTimeQuantum, timeAsString(lastActivated));
            printf("%s).\n", timeAsString(currentTime)); ++linesToDelete;
            printf("%lu to run, %lu ready, %lu blocked, %lu running, %lu finished.\n", toRunSize, readySize, blockedSize, executingSize, finishedSize); ++linesToDelete;
            
            printf("+----------+-------+-------+--------+--------+--------+--------+--------+-------+-------+-------+\n"); ++linesToDelete;
            printf("|  QUEUE   |  PID  | Prior | R Time | T Time | E Time | E Rema | W Time | Queue | WFact | T Rem |\n"); ++linesToDelete;
            printf("+----------+-------+-------+--------+--------+--------+--------+--------+-------+-------+-------+\n"); ++linesToDelete;
            
            if (toRunSize > 0) {
                printf("| To Run   |       |       |        |        |        |        |        |       |       |       |\n"); ++linesToDelete;
                
                for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    averagePriority += current->currentPriority;
                    
                    curResponse = timeDifference(currentTime, current->launched);
                    curTurnaround = timeDifference(currentTime, current->launched);
                    curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);

                    averageResponse += curResponse;
                    averageWResponse += (1.0 / current->priority) * curResponse;
                    averageTurnaround += curTurnaround;
                    averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                    averageWaitTime += curWaitTime;
                    averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                    averageExecution += current->executionTime;
                    averageRemainingExecution += current->executionTimeRemaining;
                    
                    printf("|          |       | %5.3f | +%5.2f | +%5.2f | %6.2f | %6.2f | +%5.2f | %5d |       |       |\n", current->priority, curResponse, curTurnaround, current->executionTime, current->executionTimeRemaining, curWaitTime, current->queue); ++linesToDelete;
                }
            }
            
            if (readySize > 0) {
                printf("| Ready    |       |       |        |        |        |        |        |       |       |       |\n"); ++linesToDelete;
                
                for (i = 0; i < numberOfQueues; ++i) {
                    if (!ready[i].empty()) {
                        for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
                            current = *iter;
                            
                            averagePriority += current->currentPriority;
                            
                            curResponse = timeDifference(current->firstServed, current->launched);
                            curTurnaround = timeDifference(currentTime, current->launched);
                            curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);

                            averageResponse += curResponse;
                            averageWResponse += (1.0 / current->priority) * curResponse;
                            averageTurnaround += curTurnaround;
                            averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                            averageWaitTime += curWaitTime;
                            averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                            averageExecution += current->executionTime;
                            averageRemainingExecution += current->executionTimeRemaining;

                            printf("|          | %5d | %5.3f | %6.2f | +%5.2f | %6.2f | %6.2f | +%5.2f | %5d | %5.2f | %5d |\n", current->pid, current->currentPriority, curResponse, curTurnaround, current->executionTime, current->executionTimeRemaining, curWaitTime, current->queue, waitTime[current->pid], timeRemaining[current->pid]); ++linesToDelete;
                        }
                    }
                }
            }
            
            if (blockedSize > 0) {
                printf("| Blocked  |       |       |        |        |        |        |        |       |       |       |\n"); ++linesToDelete;
                
                for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    averagePriority += current->currentPriority;
                    
                    curResponse = timeDifference(current->firstServed, current->launched);
                    curTurnaround = timeDifference(currentTime, current->launched);
                    curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);

                    averageResponse += curResponse;
                    averageWResponse += (1.0 / current->priority) * curResponse;
                    averageTurnaround += curTurnaround;
                    averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                    averageWaitTime += curWaitTime;
                    averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                    averageExecution += current->executionTime;
                    averageRemainingExecution += current->executionTimeRemaining;

                    printf("|          | %5d | %5.3f | %6.2f | +%5.2f | %6.2f | %6.2f | +%5.2f | %5d | %5.2f | %5d |\n", current->pid, current->currentPriority, curResponse, curTurnaround, current->executionTime, current->executionTimeRemaining, curWaitTime, current->queue, waitTime[current->pid], timeRemaining[current->pid]); ++linesToDelete;
                }
            }
            
            if (executingSize > 0) {
                printf("| Running  |       |       |        |        |        |        |        |       |       |       |\n"); ++linesToDelete;
                
                for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    averagePriority += current->currentPriority;
                    
                    curResponse = timeDifference(current->firstServed, current->launched);
                    curTurnaround = timeDifference(currentTime, current->launched);
                    curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);

                    averageResponse += curResponse;
                    averageWResponse += (1.0 / current->priority) * curResponse;
                    averageTurnaround += curTurnaround;
                    averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                    averageWaitTime += curWaitTime;
                    averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                    averageExecution += current->executionTime;
                    averageRemainingExecution += current->executionTimeRemaining;

                    printf("|          | %5d | %5.3f | %6.2f | +%5.2f | %6.2f | %6.2f | +%5.2f | %5d |       | %5d |\n", current->pid, current->currentPriority, curResponse, curTurnaround, current->executionTime, current->executionTimeRemaining, curWaitTime, current->queue, timeRemaining[current->pid]); ++linesToDelete;
                }
            }
            
            if (finishedSize > 0) {
                printf("| Finished |       |       |        |        |        |        |        |       |       |       |\n"); ++linesToDelete;
                
                for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
                    current = *iter;
                    
                    averagePriority += current->currentPriority;
                    
                    curResponse = timeDifference(current->firstServed, current->launched);
                    curTurnaround = timeDifference(current->ended, current->launched);
                    curWaitTime = curTurnaround - (current->executionTime - current->executionTimeRemaining);

                    averageResponse += curResponse;
                    averageWResponse += (1.0 / current->priority) * curResponse;
                    averageTurnaround += curTurnaround;
                    averageWTurnaround += (1.0 / current->priority) * curTurnaround;
                    averageWaitTime += curWaitTime;
                    averageWWaitTime += (1.0 / current->priority) * curWaitTime;
                    averageExecution += current->executionTime;
                    
                    printf("|          | %5d | %5.3f | %6.2f | %6.2f | %6.2f |        | %6.2f | %5d |       |       |\n", current->pid, current->currentPriority, curResponse, curTurnaround, current->executionTime, curWaitTime, current->queue); ++linesToDelete;
                }
            }
            
            double fullCount = toRunSize + readySize + blockedSize + executingSize + finishedSize;
            
            averagePriority /= fullCount;
            averageResponse /= fullCount;
            averageWResponse /= fullCount;
            averageTurnaround /= fullCount;
            averageWTurnaround /= fullCount;
            averageWaitTime /= fullCount;
            averageWWaitTime /= fullCount;
            averageExecution /= fullCount;
            averageRemainingExecution /= fullCount;
            
            printf("+----------+-------+-------+--------+--------+--------+--------+--------+-------+-------+-------+\n"); ++linesToDelete;
            
            printf("| AVERAGE  |       | %5.2f | %6.2f | %6.2f | %6.2f | %6.2f | %6.2f |       |       |       |\n", averagePriority, averageResponse, averageTurnaround, averageExecution, averageRemainingExecution, averageWaitTime); ++linesToDelete;
            printf("| WAVERAGE |       |       | %6.2f | %6.2f |        |        | %6.2f |       |       |       |\n", averageWResponse, averageWTurnaround, averageWWaitTime); ++linesToDelete;
            
            printf("+----------+-------+-------+--------+--------+--------+--------+--------+-------+-------+-------+\n"); ++linesToDelete;
        }
    }
}

MLFQScheduler::MLFQScheduler(bool inter, int queues, int *quantums, int *types) : Scheduler(inter), numberOfQueues(queues) {
    assert(numberOfQueues > 0);
    
    timeQuantums = new int[numberOfQueues];
    memcpy(timeQuantums, quantums, sizeof(int) * numberOfQueues);
    
    queueTypes = new int[numberOfQueues];
    memcpy(queueTypes, types, sizeof(int) * numberOfQueues);
    
    for (int i = 0; i < numberOfQueues; ++i) {
        if ((queueTypes[i] < MLFQ_QUEUE_TYPE_MIN) || (queueTypes[i] > MLFQ_QUEUE_TYPE_MAX) || ((queueTypes[i] == MLFQ_PRR) && (i > 0))) {
            fprintf(stderr, "MLFQScheduler::MLFQScheduler(%s, %d, ..., ...) - Invalid type for queue number #%d.\n", (inter ? "true" : "false"), queues, (i + 1));
            exit(-72);
        }
    }
    
    ready = new list<ProcessInfo*>[numberOfQueues]();
}

MLFQScheduler::~MLFQScheduler(void) {
    list<ProcessInfo*>::iterator iter, stopper;
    ProcessInfo *current;
    
    for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
        delete *iter;
    }
    
    for (int i = 0; i < numberOfQueues; ++i) {
        for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
            current = *iter;
            
            if (0 != kill(current->pid, SIGKILL)) {
                fprintf(stderr, "MLFQScheduler::~MLFQScheduler() - Unable to kill process #%d, error #%d.\n", current->pid, errno);
            }
            
            delete current;
        }
    }
    
    for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
        current = *iter;
        
        if (0 != kill(current->pid, SIGKILL)) {
            fprintf(stderr, "MLFQScheduler::~MLFQScheduler() - Unable to kill process #%d, error #%d.\n", current->pid, errno);
        }
        
        delete current;
    }
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        current = *iter;
        
        if (0 != kill(current->pid, SIGKILL)) {
            fprintf(stderr, "MLFQScheduler::~MLFQScheduler() - Unable to kill process #%d, error #%d.\n", current->pid, errno);
        }
        
        delete current;
    }
    
    for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
        delete *iter;
    }
    
    delete[] timeQuantums;
    delete[] queueTypes;
    delete[] ready;
}

void MLFQScheduler::setActive(bool beActive) {
    active = beActive;
    
    if (active) {
        list<ProcessInfo*>::iterator iter, stopper;
        ProcessInfo *current;
        
        gettimeofday(&lastActivated, NULL);
        
        for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
            current = *iter;
            
            if ((0 == current->launched.tv_sec) && (0 == current->launched.tv_usec)) {
                gettimeofday(&(current->launched), NULL);
            }
        }
        
        if (executing.empty()) {
            ++ticksSoFar;
            this->runNextProcess();
        }
    }
}

void MLFQScheduler::tick(void) {
    list<ProcessInfo*>::iterator iter, stopper;
    int count = 0, timeLeft;
    ProcessInfo *current;
    double currentWaitTime;
    int i, readyNSize;
    
    ++ticksSoFar;
    
    for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
        current = *iter;
        
        ready[current->queue].push_back(current);
    }
    
    blocked.clear();
    
    readyNSize = toRun.size() + executing.size() + blocked.size();
    
    for (i = 0; i < numberOfQueues; ++i) {
        readyNSize += ready[i].size();
    }

    for (i = 1; i < numberOfQueues; ++i) {
        for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
            current = *iter;
            
            currentWaitTime = waitTime[current->pid];
            
            currentWaitTime += (0.5 / (current->currentPriority * readyNSize));
            
            if (currentWaitTime >= readyNSize) {
                iter = ready[i].erase(iter);
                current->queue = i - 1;
                ready[current->queue].push_back(current);
                timeRemaining[current->pid] = timeQuantums[current->queue];
                currentWaitTime = 0.0;
            }
            
            waitTime[current->pid] = currentWaitTime;
        }
    }
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        current = *iter;
        
        current->executionTimeRemaining -= currentTimeQuantum;
        
        timeLeft = timeRemaining[current->pid] - 1;
        
        iter = executing.erase(iter);

        if (0 != kill(current->pid, SIGSTOP)) {
            fprintf(stderr, "MLFQScheduler::tick() - Unable to stop process #%d, error #%d.\n", current->pid, errno);
        } else {
            ++count;
        }
        
        if ((current->queue != (numberOfQueues - 1)) && (timeLeft <= 0)) {            
            timeRemaining.erase(current->pid);
            
            ++(current->queue);
            
            waitTime[current->pid] = 0.0;
            timeRemaining[current->pid] = timeQuantums[current->queue];
            
            ready[current->queue].push_back(current);
        } else {
            timeRemaining[current->pid] = timeLeft;
            
            if ((queueTypes[current->queue] == MLFQ_RR) || (queueTypes[current->queue] == MLFQ_PRR)) {
                ready[current->queue].push_back(current);
            } else {
                ready[current->queue].push_front(current);
            }
        }
    }
    
    while (count-- > 0) {
        this->runNextProcess();
    }
    
    if (interactive) {
        updateInteractiveDisplay();
    }
}

void MLFQScheduler::startProcess(double priority, double executionTime) {
    ProcessInfo *temp = new ProcessInfo;
    
    temp->priority = temp->currentPriority = priority;
    temp->executionTime = temp->executionTimeRemaining = executionTime;
    
    if (active) {
        gettimeofday(&(temp->launched), NULL);
    } else {
        temp->launched.tv_sec = temp->launched.tv_usec = 0;
    }
    
    toRun.push_back(temp);

    if (executing.empty()) {
        this->runNextProcess();
    }
    
    if (interactive) {
        updateInteractiveDisplay();
    }
}

vector<pid_t>* MLFQScheduler::processList(void) {
    list<ProcessInfo*>::const_iterator iter, stopper;
    vector<pid_t> *result = new vector<pid_t>;
    int i;
    
    for (i = 0; i < numberOfQueues; ++i) {
        for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
            result->push_back((*iter)->pid);
        }
    }
    
    for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
        result->push_back((*iter)->pid);
    }
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        result->push_back((*iter)->pid);
    }
    
    return result;
}

ProcessInfo* MLFQScheduler::processInfo(pid_t pid) {
    list<ProcessInfo*>::const_iterator iter, stopper;
    int i;
    
    for (i = 0; i < numberOfQueues; ++i) {
        for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
            if ((*iter)->pid == pid) {
                return copyInfo(*iter);
            }
        }
    }
    
    for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            return copyInfo(*iter);
        }
    }
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            return copyInfo(*iter);
        }
    }
    
    for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            return copyInfo(*iter);
        }
    }
    
    return NULL;
}

void MLFQScheduler::killProcess(pid_t pid) {
    list<ProcessInfo*>::const_iterator iter, stopper;
    int i;
    
    for (i = 0; i < numberOfQueues; ++i) {
        for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
            if ((*iter)->pid == pid) {
                kill(SIGKILL, pid); return;
            }
        }
    }
    
    for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            kill(SIGKILL, pid); return;
        }
    }
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            kill(SIGKILL, pid); return;
        }
    }
    
    for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
        if ((*iter)->pid == pid) {
            kill(SIGKILL, pid); return;
        }
    }
    
    fprintf(stderr, "MLFQScheduler::killProcess(%d) - Unable to find any process with that real-world PID.  Remember I can only kill processes that I am managing, not just any arbitrary ones.\n", pid);
}

void MLFQScheduler::processBlocked(pid_t pid) {
    list<ProcessInfo*>::iterator iter, stopper;
    ProcessInfo *current;
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        current = *iter;
        
        if (pid == current->pid) {
            blocked.push_back(current);
            executing.erase(iter);
            
            this->runNextProcess();
            
            if (interactive) {
                this->updateInteractiveDisplay();
            }
            
            return;
        }
    }
    
    fprintf(stderr, "MLFQScheduler::processBlocked(%d) - Cannot find an executing process with the given PID.\n", pid);
}

void MLFQScheduler::processTerminated(pid_t pid, int status) {
    list<ProcessInfo*>::iterator iter, stopper;
    ProcessInfo *selection = NULL, *current;
    bool wasActive = false;
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        current = *iter;
        
        if (current->pid == pid) {
            selection = current;
            executing.erase(iter);
            wasActive = true;
            break;
        }
    }
    
    // In theory we could stop here, since unless there's outside interference only an executing process could be killed.  But, since the user or other programs might kill a stopped process nonetheless, we need to be complete.
    
    if (!selection) {
        int i = 0;
        
        while (!selection && (i < numberOfQueues)) {
            for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
                current = *iter;
                
                if (current->pid == pid) {
                    selection = current;
                    ready[i].erase(iter);
                    break;
                }
            }
            
            ++i;
        }
        
        if (!selection) {
            for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
                current = *iter;
                
                if (current->pid == pid) {
                    selection = current;
                    blocked.erase(iter);
                    break;
                }
            }
            
            if (!selection) {
                fprintf(stderr, "MLFQScheduler::processTerminated(%d, %d) - Unable to find any record of the given process.\n", pid, status);
                return;
            }
        }
    }
    
    selection->result = status;
    
    if (0 != gettimeofday(&(selection->ended), NULL)) {
        fprintf(stderr, "MLFQScheduler::processTerminated() - Unable to get time of day, error #%d.\n", errno);
    }
    
    finished.push_back(selection);
    
    if (wasActive) {
        this->runNextProcess();
    }
    
    if (interactive) {
        updateInteractiveDisplay();
    }
}

void MLFQScheduler::outputState(ostream &out) {
    unsigned long toRunSize = toRun.size(), readySize = 0, blockedSize = blocked.size(), finishedSize = finished.size(), executingSize = executing.size(); // It's important to only call size() on a list once, where possible, as calculating the size involves iterating over the entire contents of the list.  This is incredibly slow for more than a few hundred thousand entries.
    ProcessInfo *current;
    list<ProcessInfo*>::const_iterator iter, stopper;
    int i;
    
    for (i = 0; i <= numberOfQueues; ++i) {
        readySize += ready[i].size();
    }
    
    out << toRunSize << " process" << ((toRunSize == 1) ? "" : "es") << " to run, " << readySize << " process" << ((readySize == 1) ? "" : "es") << " ready, " << blockedSize << " process" << ((blockedSize == 1) ? "" : "es") << " blocked, " << executingSize << " process" << ((executingSize == 1) ? "" : "es") << " executing, " << finishedSize << " process" << ((finishedSize == 1) ? "" : "es") << " finished.\n";

    if (toRunSize > 0) {
        out << "To run:\n";
        
        for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
            current = *iter;
            
            printf("\tPriority: %2.2f\tExecution time: %2.2f\tLaunched: %s (%s)\n", current->priority, current->executionTime, timeAsString(current->launched), timeSinceNowAsString(current->launched));
        }
    }
    
    if (readySize > 0) {
        out << "Ready:\n";
        
        for (i = 0; i < numberOfQueues; ++i) {
            printf("\tQueue %d:\n", i);
            
            for (iter = ready[i].begin(), stopper = ready[i].end(); iter != stopper; ++iter) {
                current = *iter;
                
                out << "\t\tPriority: " << current->priority << "\tExecution time: " << current->executionTime << "\tLaunched: " << timeAsString(current->launched) << " (" << timeSinceNowAsString(current->launched) << ")\tPID: " << current->pid << "\tCurrent priority: " << current->currentPriority;
                out << "\t\tFirst served: " << timeAsString(current->firstServed) << " (" << timeSinceNowAsString(current->firstServed) << ")\n";
            }
        }
    }
    
    if (blockedSize > 0) {
        out << "Blocked:\n";
        
        for (iter = blocked.begin(), stopper = blocked.end(); iter != stopper; ++iter) {
            current = *iter;
            
            out << "\tPriority: " << current->priority << "\tExecution time: " << current->executionTime << "\tLaunched: " << timeAsString(current->launched) << " (" << timeSinceNowAsString(current->launched) << ")\tPID: " << current->pid << "\tCurrent priority: " << current->currentPriority;
            out << "\tFirst served: " << timeAsString(current->firstServed) << " (" << timeSinceNowAsString(current->firstServed) << ")\n";}
    }
    
    if (executingSize > 0) {
        out << "Executing:\n";
        
        for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
            current = *iter;
            
            out << "\tPriority: " << current->priority << "\tExecution time: " << current->executionTime << "\tLaunched: " << timeAsString(current->launched) << " (" << timeSinceNowAsString(current->launched) << ")\tPID: " << current->pid << "\tCurrent priority: " << current->currentPriority;
            out << "\tFirst served: " << timeAsString(current->firstServed) << " (" << timeSinceNowAsString(current->firstServed) << ")\n";
        }
    }
    
    if (finishedSize > 0) {
        out << "Finished:\n";
        
        for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
            current = *iter;
            
            out << "\tPriority: " << current->priority << "\tExecution time: " << current->executionTime << "\tLaunched: " << timeAsString(current->launched) << " (" << timeSinceNowAsString(current->launched) << ")\tPID: " << current->pid << "\tCurrent priority: " << current->currentPriority;
            out << "\tFirst served: " << timeAsString(current->firstServed) << " (" << timeSinceNowAsString(current->firstServed) << ")";
            out << "\tEnded: " << timeAsString(current->ended) << " (" << timeSinceNowAsString(current->ended) << ")\n";
        }
    }
}

void MLFQScheduler::outputPerformance(ostream &out) {
    list<ProcessInfo*>::const_iterator iter, stopper;
    ProcessInfo *current;
    double bestTurnaround = 10000000000.0, totalTurnaround = 0.0, worstTurnaround = 0.0, bestResponse = 10000000000.0, totalResponse = 0.0, worstResponse = 0.0, currentTurnaround, currentResponse; // Note that the best in both cases is set to a massive initial value to ensure they will take on the value of the first process considered.  For assurance, the given defaults represent a little over 3 centuries of time.. if you run this program for that long, good for you.
    double bestTurnaroundWeighted = 10000000000.0, totalTurnaroundWeighted = 0.0, worstTurnaroundWeighted = 0.0, bestResponseWeighted = 10000000000.0, totalResponseWeighted = 0.0, worstResponseWeighted = 0.0, currentTurnaroundWeighted, currentResponseWeighted;
    
    int count = 0;
    
    if (!finished.empty()) {
        for (iter = finished.begin(), stopper = finished.end(); iter != stopper; ++iter) {
            ++count;
            current = *iter;
            
            currentTurnaround = (current->ended.tv_sec + (current->ended.tv_usec / 1000000.0)) - (current->launched.tv_sec + (current->launched.tv_usec / 1000000.0));
            currentResponse = (current->firstServed.tv_sec + (current->firstServed.tv_usec / 1000000.0)) - (current->launched.tv_sec + (current->launched.tv_usec / 1000000.0));
            
            currentTurnaroundWeighted = currentTurnaround * (1.0 / current->priority);
            currentResponseWeighted = currentResponse * (1.0 / current->priority);
            
            totalTurnaround += currentTurnaround;
            totalResponse += currentResponse;
            
            totalTurnaroundWeighted += currentTurnaroundWeighted;
            totalResponseWeighted += currentResponseWeighted;
            
            if (currentTurnaround < bestTurnaround) {
                bestTurnaround = currentTurnaround;
            }
            
            if (currentTurnaround > worstTurnaround) {
                worstTurnaround = currentTurnaround;
            }
            
            if (currentResponse < bestResponse) {
                bestResponse = currentResponse;
            }
            
            if (currentResponse > worstResponse) {
                worstResponse = currentResponse;
            }
            
            if (currentTurnaroundWeighted < bestTurnaroundWeighted) {
                bestTurnaroundWeighted = currentTurnaroundWeighted;
            }
            
            if (currentTurnaroundWeighted > worstTurnaroundWeighted) {
                worstTurnaroundWeighted = currentTurnaroundWeighted;
            }
            
            if (currentResponseWeighted < bestResponseWeighted) {
                bestResponseWeighted = currentResponseWeighted;
            }
            
            if (currentResponseWeighted > worstResponseWeighted) {
                worstResponseWeighted = currentResponseWeighted;
            }
        }
        
        out << count << " process" << ((count == 1) ? "" : "es") << " completed:\n\tTurnaround { Best: " << bestTurnaround << "\tWorst: " << worstTurnaround << "\tAverage: " << (totalTurnaround / count) << " }\n\tResponse { Best: " << bestResponse << "\tWorst: " << worstResponse << "\tAverage: " << (totalResponse / count) << " }\n\tWeighted turnaround { Best: " << bestTurnaroundWeighted << "\tWorst: " << worstTurnaroundWeighted << "\tAverage: " << (totalTurnaroundWeighted / count) << " }\n\tWeighted response { Best: " << bestResponseWeighted << "\tWorst: " << worstResponseWeighted << "\tAverage: " << (totalResponseWeighted / count) << " }\n";
    }
}
