//  MLFQScheduler.h
//  Scheduler Simulator
//
//  Created by Wade Tregaskis on Fri Apr 09 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef MLFQ_SCHEDULER_H
#define MLFQ_SCHEDULER_H

#include "Scheduler.h"
#include "ListAdditions.h"

#include <map>
#include <list>
using namespace std;


extern const int MLFQ_RR;
extern const int MLFQ_PRR;
extern const int MLFQ_FCFS;
extern const int MLFQ_SRJF;
extern const int MLFQ_P;


/*! @class MLFQScheduler
    @abstract MLFQ (Multi-Level Feedback Queue) Scheduler.
    @discussion The MLFQ is basically a combination of any and all other types of schedulers.  It operates one or more queues of ready processes, each of which may use a different algorithm for scheduling.  E.g. the first queue may be RR, and the second FCFS - when processes have used up their allocated time allowance for the RR queue, they fall back to the FCFS one.  Earlier queues get higher priority.  Processes are automatically bumped up a queue if they spend enough time waiting in their current one - how much time they must wait depends on their priority and the number of other processes operating.
                For details of each of the queue types supported, see the relevant documentation for the individual scheduler of that type.  Where possible the MLFQ schedulers are kept true to their individual behaviours, although for some this is not possible (e.g. the P scheduler does not operate the same because it is not allowed to change the priority of processes).
                Note that this scheduler uses a 4 + N queue model, similar to the 5-queue model used by most singular schedulers.  As such much of it's functionality, interface and code is duplicated from the FiveQueueScheduler class.  This is not easily avoidable, as there are sufficient small differences that a common superclass is not really feasable.
    */

class MLFQScheduler : public Scheduler {
protected:
    int numberOfQueues; // The number of queues to use.  This does not include the final RR queue - just the FCFS initial queues.
    int *timeQuantums, *queueTypes; // 'timeQuantums' is an array of values indicating how much many time quantums each process is given in each queue (contains 'numberOfQueues' entries.  i.e. a value of 1 indicates one time quantum.  'queueTypes' stores the type of each queue.
    list<ProcessInfo*> toRun, *ready, blocked, finished, executing; // Ready is an array of 'numberOfQueues' + 1 queues.  The last one is the RR queue into which all long-running processes end up.
    map<int, int> timeRemaining; // Contains the time remaining for each executing process.  The process PID is used as the key.
    map<int, double> waitTime; // The internal measure of wait time (in arbitrary units) which determines when (if) processes will be promoted to a higher-priority queue.  The process PID is used as the key.
    
/*! @function runNextProcess
    @abstract Instructs the receiver to locate the next process to run, and do so.
    @discussion This is where most of the guts of the MLFQ scheduler is; determining which process to run next.  This method chooses some appropriate process, depending on the MLFQ setup (queue types, location of processes within those queues, etc etc), and runs it. */
    
    virtual void runNextProcess(void);
    
/*! @function updateInteractiveDisplay
    @abstract Updates the interactive display of MLFQ scheduler operation.
    @discussion This method displays a table of the receiver's scheduled processes, alongside any additional information that is relevant.  It also makes sure to clear/overwrite the previous table, if one has been rendered.  It assumes that no other output is presented to the user inbetween invocations. */
    
    virtual void updateInteractiveDisplay(void);

public:
    MLFQScheduler(bool inter, int queues, int *quantums, int *types);
    ~MLFQScheduler(void);
    
    // Scheduler functions
    
    virtual void setActive(bool beActive);
    
    virtual void tick(void);
    
    virtual void startProcess(double priority, double executionTime);
    virtual vector<pid_t>* processList(void);
    virtual ProcessInfo* processInfo(pid_t pid);
    virtual void killProcess(pid_t pid);

    virtual void processBlocked(pid_t pid);
    virtual void processTerminated(pid_t pid, int status);
    
    virtual void outputState(ostream &out);
    virtual void outputPerformance(ostream &out);
};

#endif
