//  PScheduler.cpp
//  Scheduler Simulator
//
//  Created by Wade Tregaskis on Fri Apr 09 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "PScheduler.h"


float PScheduler::executionPriorityChange(void) {
    return epc;
}

float PScheduler::idlePriorityChange(void) {
    return ipc;
}

void PScheduler::setExecutionPriorityChange(float newEPC) {
    epc = newEPC;
}

void PScheduler::setIdlePriorityChange(float newIPC) {
    ipc = newIPC;
}

void PScheduler::runNextProcess(void) {
    if (active) {
        ProcessInfo *current, *selection = NULL;
        list<ProcessInfo*>::iterator iter, stopper;
        bool toRunChoice = true;
        
        for (iter = toRun.begin(), stopper = toRun.end(); iter != stopper; ++iter) {
            current = *iter;
            
            if (!selection || (current->currentPriority < selection->currentPriority)) {
                selection = current;
            }
        }
        
        for (iter = ready.begin(), stopper = ready.end(); iter != stopper; ++iter) {
            current = *iter;
            
            if (!selection || (current->currentPriority < selection->currentPriority)) {
                toRunChoice = false;
                selection = current;
            }
        }
        
        if (selection) {
            selection->currentPriority += epc;
            executing.push_back(selection);
            
            if (toRunChoice) {
                remove(&toRun, selection);
                
                if (0 != gettimeofday(&(selection->firstServed), NULL)) {
                    fprintf(stderr, "PScheduler::runNextProcess() - Unable to get time of day, error #%d.\n", errno);
                }
                
                selection->pid = spawnChild(selection->executionTime);
                
                if (selection->pid <= 0) {
                    fprintf(stderr, "PScheduler::runNextProcess() - Unable to spawn process, error #%d.\n", errno);
                }        
            } else {
                remove(&ready, selection);
                
                if (0 != kill(selection->pid, SIGCONT)) {
                    fprintf(stderr, "PScheduler::runNextProcess() - Unable to wake up child process #%d - error #%d.\n", selection->pid, errno);
                }
            }
        }
    }
}

PScheduler::PScheduler(bool inter, double initialEPC, double initialIPC) : FiveQueueScheduler(inter), epc(initialEPC), ipc(initialIPC) {
    // Do nothing
}

PScheduler::~PScheduler(void) {
    // Do nothing
}

void PScheduler::tick(void) {
    list<ProcessInfo*>::iterator iter, stopper;
    int count = 0;
    ProcessInfo *current;
    
    ++ticksSoFar;
    
    // Before we added blocked processes, we apply the idlePriorityChange to those that are ready.  These are the processes that have not executed during the previous time quantum - therefore their priority should be adjusted.  Processes which are currently blocked are assumed to have executed in the previous time quantum, or otherwise be excluded from idle priority adjustment, and are only added back into the ready list afterwards.
    
    for (iter = ready.begin(), stopper = ready.end(); iter != stopper; ++iter) {
        (*iter)->currentPriority += ipc;
    }
    
    ready.insert(ready.end(), blocked.begin(), blocked.end());
    blocked.clear();
    
    for (iter = executing.begin(), stopper = executing.end(); iter != stopper; ++iter) {
        current = *iter;
        
        current->executionTimeRemaining -= currentTimeQuantum;

        if (0 != kill(current->pid, SIGSTOP)) {
            fprintf(stderr, "PScheduler::tick() - Unable to stop process #%d, error #%d.\n", current->pid, errno);
        } else {
            ++count;
        }
        
        ready.push_back(current);
    }
    
    executing.clear();
    
    while (count--) {
        this->runNextProcess();
    }
    
    if (interactive) {
        this->updateInteractiveDisplay();
    }
}
