//  PScheduler.h
//  Scheduler Simulator
//
//  Created by Wade Tregaskis on Fri Apr 09 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef P_SCHEDULER_H
#define P_SCHEDULER_H

#include "FiveQueueScheduler.h"
#include "ListAdditions.h"


/*! @class PScheduler
    @abstract P (Priority) Scheduler.
    @discussion Similar to a RR scheduler, except that instead of executing processes in the same order over and over, this executes the process with the highest priority at dispatch time.  It then adjusts the priority of that process (and those that don't get selected) so that frequently executed processes drop in priority.  Over time all process priorities average out to be equal, and so the scheduler degenerates to RR.
                The idea behind the P scheduler is two-fold:  in the 'balanced' case, all processes end up at mid-priority after some sufficient time.  Thus, when a new process is submitted to the scheduler it gets either a significant initial boost or supression, depending on whether it's priority is higher or lower than average (respectively).  All the while the scheduler tends to degenerate towards a 'fair' RR mode of operation.
                The alternative approach is to operate in an 'unbalanced' case.  Generally this is done so that processes all reach the same priority after a sufficient time period, but this average priority increases slowly.  Thus, new processes even of the same [original] priority as the existing processes are treated as being lower priority, as they are newer and thus there is less need to execute them.  It could also occur conversely - the average priority could reduce over time.  This would increase general response times and provide similar operation to a foreground-batch MLQ.
    */

class PScheduler : public FiveQueueScheduler {
protected:
    float epc, ipc;
        
    // FiveQueueScheduler
    
    virtual void runNextProcess(void);

public:
/*! @function PScheduler
    @abstract Designated initialiser for the PScheduler class.
    @discussion Initialises the receiver with the initial values for EPC & IPC as given, or to the default values if not.
    @param initialEPC The initial value for EPC (Execution Priority Change).
    @param initialIPC The initial value for IPC (Idle Priority Change). */
        
    PScheduler(bool inter = false, double initialEPC = 0.1, double initialIPC = -0.01);
    ~PScheduler(void);
    
/*! @function executionPriorityChange
    @abstract Returns the current EPC.
    @result Returns the current EPC. */
    
    virtual float executionPriorityChange(void);
    
/*! @function idlePriorityChange
    @abstract Returns the current IPC.
    @result Returns the current IPC. */
    
    virtual float idlePriorityChange(void);
    
/*! @function setExecutionPriorityChange
    @abstract Sets the EPC of the receiver to a given value.
    @discussion Use this member function to adjust the EPC after initialisation.  It is possible to change it at any time and to any value you desire.  Just don't expect useful results if you go crazy.
    @param newEPC The new EPC value. */
    
    virtual void setExecutionPriorityChange(float newEPC);
    
/*! @function setIdlePriorityChange
    @abstract Sets the IPC of the receiver to a given value.
    @discussion Use this member function to adjust the IPC after initialisation.  It is possible to change it at any time and to any value you desire.  Just don't expect useful results if you go crazy.
    @param newIPC The new IPC value. */
    
    virtual void setIdlePriorityChange(float newIPC);
    
    // Scheduler
    
    virtual void tick(void);
};

#endif
