//  Scheduler.cpp
//  Scheduler Simulator
//
//  Created by Wade Tregaskis on Fri Apr 09 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "Scheduler.h"


ProcessInfo* copyInfo(ProcessInfo *info) {
    if (info) {
        ProcessInfo *temp = new ProcessInfo;
        
        memcpy(temp, info, sizeof(ProcessInfo));
        
        return temp;
    } else {
        return NULL;
    }
}

Scheduler::~Scheduler(void) {
    // Do nothing
}

pid_t spawnChild(double executionTime) {
    sigset_t oldSignalSet, blockSet;
    
    sigemptyset(&blockSet);
    sigaddset(&blockSet, SIGUSR1);
    sigaddset(&blockSet, SIGCHLD);
    sigaddset(&blockSet, SIGALRM);
    
    assert(sigprocmask(SIG_BLOCK, &blockSet, &oldSignalSet) == 0); // We need to disable these because our child inherits the signal stuff, which means it could get RT timer ticks and that kind of crap too... that would be bad.
    
    pid_t result = fork();
    
    if (result < 0) { // Fork failed
        fprintf(stderr, "Unable to spawn child process for time %f because error #%d occured.\n", executionTime, errno);
        fflush(stderr);
        exit(-100);
        return -1;
    } else if (result > 0) { // We are the parent
        assert(sigprocmask(SIG_SETMASK, &oldSignalSet, NULL) == 0);
        
        return result;
    } else { // We are the child
        struct rusage myUsage;
        double activeTime, nextNotification = 0.5;
        pid_t mypid = getpid();
        struct itimerval time;
        struct sigaction action;
        
        action.sa_handler = SIG_DFL;
        sigemptyset(&(action.sa_mask));
        action.sa_flags = 0;
        
        time.it_value.tv_sec = time.it_value.tv_usec = time.it_interval.tv_sec = time.it_interval.tv_usec = 0;
        
        assert(0 == sigaction(SIGCHLD, &action, NULL));
        assert(0 == sigaction(SIGUSR1, &action, NULL));
        assert(0 == sigaction(SIGALRM, &action, NULL));
        assert(0 == setitimer(ITIMER_REAL, &time, NULL));
        
        //printf("Child process %d spawned to use %f seconds of CPU time.\n", mypid, executionTime);
        
        while (1) {
            if (0 != getrusage(RUSAGE_SELF, &myUsage)) {
                fprintf(stderr, "Child process %d couldn't get CPU usage stats - dying.\n", mypid);
                exit(-1);
            }
            
            activeTime = (myUsage.ru_utime.tv_sec + (myUsage.ru_utime.tv_usec / 1000000.0)) + (myUsage.ru_stime.tv_sec + (myUsage.ru_stime.tv_usec / 1000000.0));
            
            if (activeTime >= executionTime) {
                //printf("Child process %d quitting after using %f seconds (aiming at %f).\n", mypid, activeTime, executionTime);
                break;
            } else if (activeTime > nextNotification) {
                //printf("Child process %d has used %f seconds so far.\n", mypid, activeTime);
                
                nextNotification += 0.5;
            }
        }
        
        exit(0);
    }
}


static Scheduler *activeScheduler = NULL;


void childBlocked(int signal, siginfo_t *info, void *data) {
    if (activeScheduler) {
        if (0 != kill(info->si_pid, SIGSTOP)) {
            fprintf(stderr, "Unable to stop (SIGSTOP) process #%d in childBlocked signal handler (i.e. the process in question just SIGUSR1'd us) - error #%d.\n", info->si_pid, errno);
        }
        
        activeScheduler->processBlocked(info->si_pid);
    }
}

void childStatusChanged(int signal, siginfo_t *info, void *data) {
    if (activeScheduler) {
        activeScheduler->processTerminated(info->si_pid, info->si_status);
    }
}

void timerTick(int signal, siginfo_t *info, void *data) {
    if (activeScheduler) {
        activeScheduler->tick();
    }
}


double currentTimeQuantum = 0.0;

int activateScheduler(Scheduler *scheduler, double timeQuantum) {
    struct itimerval timer;
    struct sigaction action;

    currentTimeQuantum = (scheduler ? timeQuantum : 0.0);
    
    sigemptyset(&(action.sa_mask)); // Block all the signals we do processing on, so that there are no issues with access to the process lists and the like
    sigaddset(&(action.sa_mask), SIGUSR1);
    sigaddset(&(action.sa_mask), SIGCHLD);
    sigaddset(&(action.sa_mask), SIGALRM);
    
    if (activeScheduler) {
        activeScheduler->setActive(false);
    }
    
    activeScheduler = scheduler;
    
    if (activeScheduler) {
        action.sa_sigaction = childStatusChanged;
        action.sa_flags = SA_SIGINFO | SA_NOCLDSTOP;
    } else {
        action.sa_handler = SIG_DFL;
        action.sa_flags = 0;
    }
    
    if (0 != sigaction(SIGCHLD, &action, NULL)) {
        fprintf(stderr, "activateScheduler() - Unable to install/remove signal handler for SIGCHLD, error #%d.\n", errno);
        return errno;
    }
    
    if (activeScheduler) {
        action.sa_sigaction = childBlocked;
        action.sa_flags = SA_SIGINFO;
    }
    
    if (0 != sigaction(SIGUSR1, &action, NULL)) {
        fprintf(stderr, "activateScheduler() - Unable to install/remove signal handler for SIGUSR1, error #%d.\n", errno);
        return errno;
    }
    
    if (activeScheduler) {
        if (timeQuantum == 0.0) {
            action.sa_handler = SIG_IGN; // SIG_DFL; // Note that if we installed the default handler, then happened to get this signal called before we could update the timer interval (below), the default behaviour is to terminate us.  So, even though we may well want the default behaviour if we're disabling scheduling (scheduler == NULL), we can't do so safely in the event that it has already been previously enabled.  Not without a fair bit of extra work, anyway.
            action.sa_flags = 0;
        } else {
            action.sa_sigaction = timerTick;
            action.sa_flags = SA_SIGINFO;
        }
    }
    
    if (0 != sigaction(SIGALRM, &action, NULL)) {
        fprintf(stderr, "activateScheduler() - Unable to install signal handler for SIGALRM, error #%d.\n", errno);
        return errno;
    }
    
    timer.it_interval.tv_sec = (int)floor(timeQuantum);
    timer.it_interval.tv_usec = (int)round((timeQuantum - timer.it_interval.tv_sec) * 1000000.0);
    
    if (0.0 == timeQuantum) {
        timer.it_value.tv_sec = timer.it_value.tv_usec = 0;
    } else {
        timer.it_value.tv_sec = timer.it_interval.tv_sec;
        timer.it_value.tv_usec = timer.it_interval.tv_usec;
    }
    
    if (0 != setitimer(ITIMER_REAL, &timer, NULL)) {
        fprintf(stderr, "activateScheduler() - Unable to configure real-time timer, error #%d.\n", errno);
        return errno;
    }
    
    if (activeScheduler) {
        activeScheduler->setActive(true);
    }
    
    return 0;
}
