//  Scheduler.h
//  Scheduler Simulator
//
//  Created by Wade Tregaskis on Fri Apr 09 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef SCHEDULER_H
#define SCHEDULER_H

#include "TimeUtils.h"

#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include <signal.h>

#include <vector>
#include <iostream>
using namespace std;


/*! @struct ProcessInfo
    @abstract Contains the information for one active, to-run or finished process.
    @discussion Just a standard data container, containing all the relevant information for a process in any of it's possible states.
    @field priority The initial priority of the process.  This will not be changed by any schedulers, and most likely not used by them after the process is launched, either.  So changing it at any particular point may not have any effect.
    @field executionTime The approximate amount of execution time the process thinks it needs.  This is not a solid value - it may be over or undershot, grossly in some cases.
    @field launched The time at which the process was launched.
    @field queue The queue in which the process currently resides (for MLFQ scheduler, or similar).
    @field pid The real-world PID of the process, if it is running, otherwise this is 0 by default.
    @field currentPriority The current priority of the process at the current point in time.  This may or may not be both used and changed by the scheduler.
    @field firstServed The time at which the process first received CPU time.
    @field executionTimeRemaining An approximation of how much CPU the process still requires, based on it's estimated executionTime and the amount of time it has been given so far.  This is generally optimistic, if anything.
    @field result The result returned by the process when it exited.
    @field ended The time at which the process terminated. */

typedef struct {
    // To be executed
    float priority; // The initial user-defined priority of the process
    double executionTime; // The amount of time required by the process
    struct timeval launched; // The time when the process was launched
    
    // Ready
    int queue; // The ID of the queue in which the process is currently placed (if applicable)
    
    // Running
    pid_t pid; // The pid of the executing or once-executed process
    float currentPriority; // The current priority of the process (some schedulers may adjust this over time)
    struct timeval firstServed; // The time when the process first received CPU time
    double executionTimeRemaining; // The approximate remaining execution time, as according to the scheduler - note that the child is quite free to ignore this.
    
    // Finished
    int result; // The result the process returned when it terminated, or -1 if it was killed
    struct timeval ended; // The time the process terminated
} ProcessInfo;


/*! @function copyInfo
    @abstract Returns a new copy of the given ProcessInfo struct.
    @discussion It is not a good idea for a scheduler to give an outside entity direct access to the actual values it uses, since they are not protected in any way.  Thus it is often necessary to copy the info and return that instead.  This is what this function does.
     @param info The information to return a copy of.
     @result A copy of the given info, or NULL if the given info is NULL.  The returned ProcessInfo is owned by the caller and must be managed as such. */

ProcessInfo* copyInfo(ProcessInfo *info);


/*! @class Scheduler
    @abstract Abstract superclass for Schedulers.
    @discussion This abstract superclass sets out the basic interface for all schedulers.  It provides process control functionality for 'end-users', i.e. the ability to add/kill jobs, list currently running jobs, etc.  It also provides event hooks for being notified of process events (kills, stops, continues, etc), as triggered from the appropriate signals or whatever else.  Finally, it offers various output functions for displaying the current state of the scheduler (and the processes it contains), as well as for performing performance analysis of the scheduler, in it's current state.
    */

class Scheduler {
protected:
    struct timeval lastActivated; // The last time the receiver was activated, as in, had active set to true (not the last time it ticked or anything)
    int ticksSoFar; // The number of ticks the receiver has received since it's creation (including alternative active periods to the current, if any).
    bool interactive, active; // Not to be confused with each other... only 'active' schedulers are permitted to run or continue processes, while 'interactive' schedulers should display an appropriate table of processes during execution.
    
public:
/*! @function Scheduler
    @abstract Designated initialiser for the Scheduler class.
    @discussion Initialises the receiver with the given parameter (if any, otherwise defaults to false).
    @param inter True if the receiver should be interactive, false otherwise.  An interactive scheduler displays a regularly updated table of processes and other information.  A non-interactive scheduler does not such thing (and requires the use of it's various informational methods in order to get useful output). */
        
    Scheduler(bool inter = false) : ticksSoFar(0), interactive(inter), active(false) {}
    virtual ~Scheduler(void); // Just setup a virtual destructor.. dumbass C++ runtime should handle inherited destructors properly to start with, but no, apparently that's too much to ask.
    
    // Scheduler Control Functions
    
/*! @function setActive
    @abstract Turn the receiver on or off.
    @discussion A scheduler must be active in order to dispatch processes to the virtual CPU(s).  By default all schedulers are created inactive, to allow appropriate setup and initialisation before they are used.  Thus, you must sooner or later call this member function to active the receiver and allow it to perform it's duties.  Note that de-activating the receiver will not immediately terminate any running processes - they will be left unmanaged and have to run to completion, or be picked up again when the scheduler is made active once more.
    @param beActive True if the scheduler should become active, false otherwise.  Asking an active scheduler to activate has no effect, and likewise for asking a non-active one to deactivate. */
    
    virtual void setActive(bool beActive) = 0;
    
    // Runloop Function
    
/*! @function tick
    @abstract Regularly called on the active scheduler to allow for pre-emptive operation.
    @discussion Pre-emptive schedulers need to be awakened regularly so as to pause running processes and re-schedule, and that is what this member function is provided for.  It is called regularly at intervals defined when activating the scheduler. */
    
    virtual void tick(void) = 0; // Called at regular intervals (the scheduling quantum) to allow the receiver to perform pre-emptive scheduling, if necessary.
    
    // Process Control Functions
/*! @function startProcess
    @abstract Starts a new process scheduled by the receiver.
    @discussion The new process may not be mapped to a real-world process immediately - most likely it will be placed in a to-run queue, where it exists only in-potentia.  Once the scheduler dispatches it for it's first CPU quantum a real-world process will be created for it.
    @param priority The priority for the new process (scheduler dependent, but by convention usually a value between 0.0 and 1.0, from highest to lowest respectively).
    @param executionTime The approximate execution time the process should require. */
    
    virtual void startProcess(double priority, double executionTime) = 0;

/*! @function processList
    @abstract Returns a list of all processes managed by the receiver.
    @discussion The returned list contains the real-world PID for each active process being managed by the receiver.  Finished and processes still to-run are not considered active - only thus executing, ready or blocked.  You can obtain more detailed information about the processes using the processInfo member function.
    @result Returns a list of real-world PID's corresponding to the active processes.  This list is owned by the caller and should be managed as such. */
    
    virtual vector<pid_t>* processList(void) = 0;
    
/*! @function processInfo
    @abstract Returns detailed information about the a process with the given real-world PID.
    @discussion Returns a non-critical copy of the info for the given process, if any.  The process in question may be any process that is running or has run (i.e. finished), but cannot be a to-run process since they do not have PID's.
    @result Returns a copy of the ProcessInfo for the given PID, or NULL if no such process is found.  The caller owns the returned ProcessInfo and is responsible for managing it appropriately. */
    
    virtual ProcessInfo* processInfo(pid_t pid) = 0;
    
/*! @function killProcess
    @abstract Instructs the receiver to kill the process with the given PID.
    @discussion The real-world process with the given PID is killed immediately.  The receiver invariably handles this in the same way as for when the child process dies through any other means - i.e. it waits for the processTerminated signal, and then reschedules as appropriate.
    @param pid The real-world PID of the process to kill. */
    
    virtual void killProcess(pid_t pid) = 0;
    
    // Process Management Events
/*! @function processBlocked
    @abstract Sent to the receiver to indicate that the given real-world process has blocked.
    @discussion This member function should generally never be called by outside code.  It indicates that the real-world process of the given PID has indicated a simulated block condition.  The receiver should handle this as appropriate.  By the time this member function is called, the process in question has already been paused (as expected to simulate blocking behaviour). */
    
    virtual void processBlocked(pid_t pid) = 0; // Called when a child process simulates a block by SIGUSR1'ing the parent.  The given process it automatically stopped.
    
/*! @function processTerminated
    @abstract Sent to the receiver to indicate that the given real-world process has terminated (for whatever reason).
    @discussion This member function should generally never be called by outside code.  It indicates that the real-world process of the given PID has terminated.  The receiver should handle this as appropriate. */
    
    virtual void processTerminated(pid_t pid, int status) = 0; // Called when a child process terminates (for whatever reason).
    
    // Output and Analysis Functions
/*! @function outputState
    @abstract Outputs a description of the receiver's current state to the given ostream.
    @discussion Warning: the output may be very long for any serious scheduler.  The output may contain anything the scheduler feels appropriate.  Usually this is a list of processes sorted into their appropriate queues, with all their details included.
    @param out The ostream to output the state description to. */
    
    virtual void outputState(ostream &out) = 0;

/*! @function outputPerformance
    @abstract Outputs an analysis of the receiver's performance to date, to the given ostream.
    @discussion The performance analysis is taken over all finished processes within the receiver - performance data cannot be evaluated for currently executing processes.  The resulting output may be rather large.
    @param out The ostream to output the analysis results to. */
    
    virtual void outputPerformance(ostream &out) = 0;
};


/*! @function spawnChild
    @abstract Spawns a child process to consume approximately the given amount of CPU time.
    @discussion The child process spawned will consume at least the given number of seconds of real CPU time (including system time) before quitting automatically.  It does nothing more.
    @param executionTime The number of seconds of CPU time for the child process to use before terminating.
    @result Returns the real-world PID of the new child, or 0 if an error occurs. */

pid_t spawnChild(double executionTime);


extern double currentTimeQuantum; // The time quantum of the currently active scheduler.  DO NOT CHANGE THIS.

/*! @function activateScheduler
    @abstract Selects the given scheduler as the currently executing one, by installing and configuring the appropriate signal handlers.
    @discussion Only one scheduler may be simulated at a time, as it needs to make use of the Unix signalling functionality to manage child processes, which does not provide a convenient way of determining which signals should be passed to which schedulers.  Consequently, successive calls to this function will replace the existing scheduler, if any.  It is the responsability of the caller to ensure the replaced scheduler(s) clean up after themselves properly, and that their managed processes do not cause further signals to the parent.
    @param scheduler The scheduler to use.
    @param timeQuantum The pre-emptive time quantum to use - i.e. the number of seconds between ticks to the scheduler.  If this is 0 the scheduler will receive no ticks.  Ticks are only necessary for pre-emptive schedulers - other schedulers will ignore them.
    @result 0 if the scheduler is installed successfully, otherwise returns the appropriate error code.  These error codes are those returned by the normal system calls sigaction() and so forth, so you can determine their meaning from man intro. */

int activateScheduler(Scheduler *scheduler, double timeQuantum);

#endif
