//  main.cpp
//  Scheduler Simulator
//
//  Created by Wade Tregaskis on Tue Apr 06 2004.
//
//  Copyright (c) 2004, Wade Tregaskis.  All rights reserved.
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//    * Neither the name of Wade Tregaskis nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include <unistd.h>

#include "FCFSScheduler.h"
#include "FCFSSScheduler.h"
#include "RRScheduler.h"
#include "PRRScheduler.h"
#include "SJFScheduler.h"
#include "SRJFScheduler.h"
#include "PScheduler.h"
#include "MLFQScheduler.h"

#include <iostream>
using namespace std;


#define RANDOM(min, max) ((min) + (((max) - (min)) * ((float)random() / RAND_MAX)))


int main(int argc, char *argv[]) {
    if ((argc < 3) || (strcasecmp(argv[1], "-h") == 0) || (strcasecmp(argv[1], "--help") == 0)) {
        printf("Usage: %s SCHED QUANT [PARAMS]\n\n\tSCHED\tThe scheduler to use, one of:\n\t\t\tRR - Round Robin\n\t\t\tPRR - Priority Round Robin\n\t\t\tSJF - Shortest Job First\n\t\t\tSRJF - Shortest Remaining Job First\n\t\t\tMLFQ - Multi-Level Feedback Queue\n\t\t\tFCFS - First Come First Served\n\t\t\tFCFSS - First Come First Served Strict\n\t\t\tP - Priority\n\tQUANT\tThe time quantum to use for pre-emptive schedulers, in seconds.\n\tPARAMS\tOptional parameters as required by the particular scheduler selected.\n", argv[0]);
        exit(1);
    } else {
        Scheduler *scheduler;
        double quantum = strtod(argv[2], NULL);
        
        if (quantum <= 0.0) {
            fprintf(stderr, "Invalid quantum size - must be greater than zero.\n");
            exit(-4);
        }
        if (strcasecmp(argv[1], "RR") == 0) {
            scheduler = new RRScheduler(true);
        } else if (strcasecmp(argv[1], "PRR") == 0) {
            scheduler = new PRRScheduler(true);
        } else if (strcasecmp(argv[1], "SJF") == 0) {
            scheduler = new SJFScheduler(true);
        } else if (strcasecmp(argv[1], "SRJF") == 0) {
            scheduler = new SRJFScheduler(true);
        } else if (strcasecmp(argv[1], "MLFQ") == 0) {
            int numberOfQueues, *quantums, *types;
            
            if ((argc < 4) || ((numberOfQueues = (int)strtol(argv[3], NULL, 10)) < 1) || (((numberOfQueues * 2) + 4) != argc)) {
                printf("Usage: %s MLFQ %s COUNT VALUE TYPE [VALUE TYPE...]\n\n\tCOUNT\tThe number of queues to use.\n\tVALUE\tThe number of time quantums to give each process in the represented queue.\n\tTYPE\tThe type of queue, one of PRR, RR, SRJF, P or FCFS.  (Note: PRR is only valid for the first queue)\n", argv[0], argv[2]);
                exit(3);
            }
            
            quantums = new int[numberOfQueues];
            types = new int[numberOfQueues];
            
            for (int i = 0; (i / 2) < numberOfQueues; i += 2) {
                quantums[i / 2] = strtol(argv[4 + i], NULL, 10);
                
                if (quantums[i / 2] < 1) {
                    fprintf(stderr, "Invalid number of quantums for queue #%d.\n", i + 1);
                    exit(-2);
                }
                
                if (strcasecmp(argv[5 + i], "RR") == 0) {
                    types[i / 2] = MLFQ_RR;
                } else if (strcasecmp(argv[5 + i], "PRR") == 0) {
                    types[i / 2] = MLFQ_PRR;
                } else if (strcasecmp(argv[5 + i], "FCFS") == 0) {
                    types[i / 2] = MLFQ_FCFS;
                } else if (strcasecmp(argv[5 + i], "SRJF") == 0) {
                    types[i / 2] = MLFQ_SRJF;
                } else if (strcasecmp(argv[5 + i], "P") == 0) {
                    types[i / 2] = MLFQ_P;
                } else {
                    fprintf(stderr, "Invalid type for queue #%d.\n", (i / 2) + 1);
                    exit(-28);
                }
                
                //printf("Queue #%d type = %d.\n", (i / 2) + 1, types[i / 2]);
            }
            
            scheduler = new MLFQScheduler(true, numberOfQueues, quantums, types);
        } else if (strcasecmp(argv[1], "FCFS") == 0) {
            scheduler = new FCFSScheduler(true);
        } else if (strcasecmp(argv[1], "FCFSS") == 0) {
            scheduler = new FCFSSScheduler(true);
        } else if (strcasecmp(argv[1], "P") == 0) {
            if (argc != 5) {
                printf("Usage: %s P %s EXEC IDLE\n\n\tEXEC\tThe amount by which to adjust the priority of tasks each time they receive 1 time quantum of execution time.\n\tIDLE\tThe amount by which to adjust the priority of each process waiting in the ready queue.\n", argv[0], argv[2]);
                exit(2);
            }
            
            scheduler = new PScheduler(true, strtod(argv[3], NULL), strtod(argv[4], NULL));
        } else {
            fprintf(stderr, "Scheduler \"%s\" not recognised - see '%s -h' for usage notes.\n", argv[1], argv[0]);
            exit(2);
        }
        
        int result, count = 0;
        vector<pid_t> *processList;
        sigset_t oldSignalSet, blockSet;
        float time;
        int priority;
        
        sigemptyset(&blockSet);
        sigaddset(&blockSet, SIGUSR1);
        sigaddset(&blockSet, SIGCHLD);
        sigaddset(&blockSet, SIGALRM);
        
        //cout << "Scheduler pid = " << getpid() << '.' << endl;
        while (1) {
            ++count;
            
            result = scanf("%f %d", &time, &priority);
            
            if (EOF == result) {
                break;
            } else if (result < 2) {
                if (result == 0) {
                    fprintf(stderr, "Unable to read execution time for process #%d.\n", count);
                } else {
                    fprintf(stderr, "Unable to read priority for process #%d.\n", count);
                }
                
                exit(-19);
            } else {
                scheduler->startProcess(1.0 / priority, time);
            }
        }
        
        /*for (i = 0; i < 10; ++i) {
            scheduler->startProcess(RANDOM(0.0, 1.0), RANDOM(1.0, 5.0));
        }*/
        
        activateScheduler(scheduler, quantum);
        
        while (1) {
            sleep(1);
            
            assert(sigprocmask(SIG_BLOCK, &blockSet, &oldSignalSet) == 0); // Prevent any signals occuring while we're accessing the process lists
            
            processList = scheduler->processList();
            
            assert(sigprocmask(SIG_SETMASK, &oldSignalSet, NULL) == 0); // Reset the signal mask to resume normal operation.  Note that the returned list is entirely ours - it does refer to any internal data of the scheduler

            assert(processList);
            
            if (processList->empty()) {
                break;
            }
            
            delete processList;
        }
        
        //scheduler->outputState(cout);
        //scheduler->outputPerformance(cout);
        
        activateScheduler(NULL, 0.0); // Switch the scheduler off
        delete scheduler;
        delete processList;
        
        return 0;
    }
}
